#ifndef _CPAR_H
#define _CPAR_H

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <real_type/real_type.h>
#ifndef _HAVE_UINT
#   define _HAVE_UINT
    typedef unsigned int uint;
#endif

#define CPAR_MAXIMUM_STRING_LENGTH 80
#define CPAR_FUNCTION_NAME __FUNCTION__

// Parameter add list modes
// Add normal key-value pair
#define CPAR_ADD_PARAMETERS_NORMAL    10000
// Merge a parameter table
#define CPAR_ADD_PARAMETERS_MERGE     10001
// Transfer a parameter from a different list to this list with a new key
#define CPAR_ADD_PARAMETERS_TRANSFER  10002
// End of parameters list
#define CPAR_ADD_PARAMETERS_END       10003

// Merge modes - Values sent to _cpar_merge()
#define CPAR_MERGE_NO_OVERWRITE_MODE   0
#define CPAR_MERGE_NEW_MODE            1
#define CPAR_MERGE_OVERWRITE_MODE      2
#define CPAR_MERGE_UPDATE_MODE         3

// CPAR basic types
#define CPAR_BASIC_TYPES                15
#define CPAR_UNKNOWN                     0
#define CPAR_BOOL                        1
#define CPAR_INT                         2
#define CPAR_UINT                        3
#define CPAR_DOUBLE                      4
#define CPAR_REAL                        5
#define CPAR_STRING                      6
#define CPAR_BOOL_POINTER                7
#define CPAR_CHAR_POINTER                8
#define CPAR_INT_POINTER                 9
#define CPAR_UINT_POINTER               10
#define CPAR_DOUBLE_POINTER             11
#define CPAR_REAL_POINTER               12
#define CPAR_CPAR_POINTER               13
#define CPAR_GPOINTER                   14
#define CPAR_GPOINTER_TYPECAST(var)  gpointer var

// Type names (created in cpar.c)
extern const char *cpar_type_names[CPAR_BASIC_TYPES];

// Custom gpointer types are a way to type-check pointers that are passed
// through gpointers.  The type in a CPAR_GET*() or CPAR_ASSIGN*() function
// must match the type used when adding to the parameter table.
// See the documentation for how to create custom pointer types.
// This is the index of the first custom type:
#define CPAR_CUSTOM_GPOINTER_TYPE       15
#define CPAR_MAXIMUM_CUSTOM_TYPES       20
// Set variables that will hold a custom type index to this value at program start:
#define CPAR_TYPE_UNINITIALIZED   UINT_MAX

// Storage types
// See the documentation for more information on storage types
#define CPAR_STORE_BY_REFERENCE         20002
#define CPAR_STORE_BY_VALUE_SIMPLE      20003
#define CPAR_STORE_BY_VALUE_COMPLEX     20004
#define CPAR_STORE_UNDEFINED            20005

// Complex store-by-value parameter have to tell cpar how to allocate, copy, and free their type
// You must provide functions with these prototypes:
#define CPAR_COMPLEX_ALLOCATE_PROTOTYPE  gpointer (*cpar_custom_allocate_function)(void)
#define CPAR_COMPLEX_COPY_PROTOTYPE  void (*cpar_custom_copy_function)(gpointer to, gpointer from)
#define CPAR_COMPLEX_FREE_PROTOTYPE  void (*cpar_custom_free_function)(gpointer)

// A union is used so that several different types can be stored in the parameter struct below
union value_union {
    bool bool_value;
    int int_value;
    uint uint_value;
    double double_value;
    real real_value;
    bool *bool_pointer_value;
    char *char_pointer_value;
    int *int_pointer_value;
    uint *uint_pointer_value;
    double *double_pointer_value;
    real *real_pointer_value;
    struct cpar_table_struct **cpar_pointer_value;
    gpointer gpointer_value;
  };

// Each parameter is stored in the hash as a pointer to a structure of this type:
struct cpar_parameter {
    uint type;
    bool is_seen;  /* set to true if this parameter has been seen */
    bool is_default;  /* set to true if this parameter is a default parameter */
    uint storage;  // See storage types above (only for gpointer)
    uint size;     // Size (only for simple store by value of a gpointer)
    CPAR_COMPLEX_ALLOCATE_PROTOTYPE;
    CPAR_COMPLEX_COPY_PROTOTYPE;
    CPAR_COMPLEX_FREE_PROTOTYPE;
    union value_union value;
  };
typedef struct cpar_parameter cpar_parameter_struct;

// The cpar master structure
struct cpar_table_struct {
  uint magic_key;
  // The pointer_type_initialization variables below aren't used yet, but they could be used
  // to store where each custom type was initialized
  char pointer_type_initialization_file[CPAR_MAXIMUM_CUSTOM_TYPES][CPAR_MAXIMUM_STRING_LENGTH];
  uint pointer_type_initialization_line[CPAR_MAXIMUM_CUSTOM_TYPES];
  GHashTable *hash_table;
};
typedef struct cpar_table_struct *cpar_table;


// CPAR_CREATE_PARAMETERS() is used inside the parameter list, and thus
// does not end with a semicolon
#define CPAR_CREATE_PARAMETERS(...) \
        _cpar_create_cpar_table(__FILE__, __LINE__, __VA_ARGS__, CPAR_ADD_PARAMETERS_END)
#define CPAR_CREATE_EMPTY_PARAMETERS() \
        _cpar_create_empty_cpar_table(__FILE__, __LINE__)
cpar_table * _cpar_create_cpar_table(char *file_name, uint file_line_number, uint first_add_mode, ...);
cpar_table * _cpar_create_empty_cpar_table(char *file_name, uint file_line_number);
cpar_parameter_struct *
_cpar_create_cpar_parameter_struct(char *file, uint line, char *key, gpointer pointer_to_value, cpar_parameter_struct *info);

// Each additional parameter in a variable parameter list created using
// CPAR_CREATE_PARAMETERS should be added with one of the following macros.
// The advantage of the specific macros for each type is that they cast the value for
// every value, meaning no wrongly-cast values in the variable parameters list.
// The advantage of the generic CPAR_ADD_GENERIC() macro is that the type can change at
// runtime.  All add a NULL to the end of each new parameter so that cpar.c can catch
// alignment errors Make sure to cast all reals if the real type is long double!

// Add
#define CPAR_ADD_GENERIC(key, type, value)  CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add                   (__FILE__, __LINE__, key, type, (gpointer)(&(value))), NULL
#define CPAR_ADD_BOOL(key, value)           CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_bool              (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_INT(key, value)            CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_int               (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_UINT(key, value)           CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_uint              (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_DOUBLE(key, value)         CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_double            (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_REAL(key, value)           CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_real              (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_STRING(key, value)         CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_string            (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_BOOL_POINTER(key, value)   CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_bool_pointer      (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_CHAR_POINTER(key, value)   CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_char_pointer      (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_INT_POINTER(key, value)    CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_int_pointer       (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_UINT_POINTER(key, value)   CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_uint_pointer      (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_DOUBLE_POINTER(key, value) CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_double_pointer    (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_REAL_POINTER(key, value)   CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_real_pointer      (__FILE__, __LINE__, key, value), NULL
#define CPAR_ADD_CPAR_POINTER(key, value)   CPAR_ADD_PARAMETERS_NORMAL, key, \
        _cpar_add_cpar_pointer      (__FILE__, __LINE__, key, value), NULL
cpar_parameter_struct * _cpar_add               (char *file, uint line, char *key, 
                                                 uint type, gpointer value);
cpar_parameter_struct * _cpar_add_bool          (char *file, uint line, char *key, bool value);
cpar_parameter_struct * _cpar_add_int           (char *file, uint line, char *key, int value);
cpar_parameter_struct * _cpar_add_uint          (char *file, uint line, char *key, uint value);
cpar_parameter_struct * _cpar_add_double        (char *file, uint line, char *key, double value);
cpar_parameter_struct * _cpar_add_real          (char *file, uint line, char *key, real value);
cpar_parameter_struct * _cpar_add_string        (char *file, uint line, char *key, char *value);
cpar_parameter_struct * _cpar_add_bool_pointer  (char *file, uint line, char *key, bool *value);
cpar_parameter_struct * _cpar_add_char_pointer  (char *file, uint line, char *key, char *value);
cpar_parameter_struct * _cpar_add_int_pointer   (char *file, uint line, char *key, int *value);
cpar_parameter_struct * _cpar_add_uint_pointer  (char *file, uint line, char *key, uint *value);
cpar_parameter_struct * _cpar_add_double_pointer(char *file, uint line, char *key, double *value);
cpar_parameter_struct * _cpar_add_real_pointer  (char *file, uint line, char *key, real *value);
cpar_parameter_struct * _cpar_add_cpar_pointer  (char *file, uint line, char *key, cpar_table *value);

// Add generic (gpointer) types
#define CPAR_ADD_GPOINTER(key, gpointer_type, value) \
    CPAR_ADD_PARAMETERS_NORMAL, key, _cpar_add_gpointer(__FILE__, __LINE__, key, gpointer_type, \
                                                        (gpointer)(value)), NULL
#define CPAR_ADD_SV_GPOINTER(key, gpointer_type, value, size) \
    CPAR_ADD_PARAMETERS_NORMAL, key, _cpar_add_sv_gpointer(__FILE__, __LINE__, key, gpointer_type, \
                                                           (gpointer)(value), size), NULL
#define CPAR_ADD_CV_GPOINTER(key, gpointer_type, value, allocate_func, copy_func, free_func) \
    CPAR_ADD_PARAMETERS_NORMAL, key, \
    _cpar_add_cv_gpointer(__FILE__, __LINE__, key, gpointer_type, \
                          (gpointer)(value), allocate_func, copy_func, free_func), NULL
cpar_parameter_struct * 
    _cpar_add_gpointer   (char *file, uint line, char *key, uint gpointer_type, gpointer value);
cpar_parameter_struct * 
    _cpar_add_sv_gpointer(char *file, uint line, char *key, uint gpointer_type, gpointer value, uint size);
cpar_parameter_struct * 
    _cpar_add_cv_gpointer(char *file, uint line, char *key, uint gpointer_type, gpointer value,
        CPAR_COMPLEX_ALLOCATE_PROTOTYPE, CPAR_COMPLEX_COPY_PROTOTYPE, CPAR_COMPLEX_FREE_PROTOTYPE);

// Add-merge
#define CPAR_ADD_MERGE(vp)              CPAR_ADD_PARAMETERS_MERGE, CPAR_MERGE_NO_OVERWRITE_MODE, vp, NULL
#define CPAR_ADD_MERGE_OVERWRITE(vp)    CPAR_ADD_PARAMETERS_MERGE, CPAR_MERGE_OVERWRITE_MODE,    vp, NULL
#define CPAR_ADD_MERGE_NEW(vp)          CPAR_ADD_PARAMETERS_MERGE, CPAR_MERGE_NEW_MODE,          vp, NULL
#define CPAR_ADD_MERGE_UPDATE(vp)       CPAR_ADD_PARAMETERS_MERGE, CPAR_MERGE_UPDATE_MODE,       vp, NULL

// Add-transfer
#define CPAR_ADD_TRANSFER(new_key, old_vp, old_key) \
    CPAR_ADD_PARAMETERS_TRANSFER, CPAR_MERGE_NO_OVERWRITE_MODE, new_key, old_vp, old_key, NULL
#define CPAR_ADD_TRANSFER_OVERWRITE(new_key, old_vp, old_key) \
    CPAR_ADD_PARAMETERS_TRANSFER, CPAR_MERGE_OVERWRITE_MODE,    new_key, old_vp, old_key, NULL
#define CPAR_ADD_TRANSFER_NEW(new_key, old_vp, old_key) \
    CPAR_ADD_PARAMETERS_TRANSFER, CPAR_MERGE_NEW_MODE,          new_key, old_vp, old_key, NULL
#define CPAR_ADD_TRANSFER_UPDATE(new_key, old_vp, old_key) \
    CPAR_ADD_PARAMETERS_TRANSFER, CPAR_MERGE_UPDATE_MODE,       new_key, old_vp, old_key, NULL

// Assign
#define CPAR_ASSIGN_BOOL(vp, key, value) \
    _cpar_assign_parameter_bool             (__FILE__, __LINE__, vp, key, value, CPAR_BOOL)
#define CPAR_ASSIGN_INT(vp, key, value) \
    _cpar_assign_parameter_int              (__FILE__, __LINE__, vp, key, value, CPAR_INT)
#define CPAR_ASSIGN_UINT(vp, key, value) \
    _cpar_assign_parameter_uint             (__FILE__, __LINE__, vp, key, value, CPAR_UINT)
#define CPAR_ASSIGN_DOUBLE(vp, key, value) \
    _cpar_assign_parameter_double           (__FILE__, __LINE__, vp, key, value, CPAR_DOUBLE)
#define CPAR_ASSIGN_REAL(vp, key, value) \
    _cpar_assign_parameter_real             (__FILE__, __LINE__, vp, key, value, CPAR_REAL)
#define CPAR_ASSIGN_STRING(vp, key, value) \
    _cpar_assign_parameter_string           (__FILE__, __LINE__, vp, key, value, CPAR_STRING)
#define CPAR_ASSIGN_BOOL_POINTER(vp, key, value) \
    _cpar_assign_parameter_bool_pointer     (__FILE__, __LINE__, vp, key, value, CPAR_BOOL_POINTER)
#define CPAR_ASSIGN_CHAR_POINTER(vp, key, value) \
    _cpar_assign_parameter_char_pointer     (__FILE__, __LINE__, vp, key, value, CPAR_CHAR_POINTER)
#define CPAR_ASSIGN_INT_POINTER(vp, key, value) \
    _cpar_assign_parameter_int_pointer      (__FILE__, __LINE__, vp, key, value, CPAR_INT_POINTER)
#define CPAR_ASSIGN_UINT_POINTER(vp, key, value) \
    _cpar_assign_parameter_uint_pointer     (__FILE__, __LINE__, vp, key, value, CPAR_UINT_POINTER)
#define CPAR_ASSIGN_DOUBLE_POINTER(vp, key, value) \
    _cpar_assign_parameter_double_pointer   (__FILE__, __LINE__, vp, key, value, CPAR_DOUBLE_POINTER)
#define CPAR_ASSIGN_REAL_POINTER(vp, key, value) \
    _cpar_assign_parameter_real_pointer     (__FILE__, __LINE__, vp, key, value, CPAR_REAL_POINTER)
#define CPAR_ASSIGN_CPAR_POINTER(vp, key, value) \
    _cpar_assign_parameter_cpar_pointer     (__FILE__, __LINE__, vp, key, value, CPAR_CPAR_POINTER)
#define CPAR_ASSIGN_GPOINTER(vp, key, gpointer_type, value) \
    _cpar_assign_parameter_gpointer         (__FILE__, __LINE__, vp, key, (gpointer)(value), gpointer_type)
bool _cpar_assign_parameter_bool            (char *file, uint line, cpar_table *vp, 
                                                     char *key, bool *, uint type);
bool _cpar_assign_parameter_int             (char *file, uint line, cpar_table *vp, 
                                                     char *key, int *, uint type);
bool _cpar_assign_parameter_uint            (char *file, uint line, cpar_table *vp, 
                                                     char *key, uint *, uint type);
bool _cpar_assign_parameter_double          (char *file, uint line, cpar_table *vp, 
                                                     char *key, double *, uint type);
bool _cpar_assign_parameter_real            (char *file, uint line, cpar_table *vp, 
                                                     char *key, real *, uint type);
bool _cpar_assign_parameter_string          (char *file, uint line, cpar_table *vp, 
                                                     char *key, char **, uint type);
bool _cpar_assign_parameter_bool_pointer    (char *file, uint line, cpar_table *vp, 
                                                     char *key, bool **, uint type);
bool _cpar_assign_parameter_char_pointer    (char *file, uint line, cpar_table *vp,
                                                     char *key, char **, uint type);
bool _cpar_assign_parameter_int_pointer     (char *file, uint line, cpar_table *vp,
                                                     char *key, int **, uint type);
bool _cpar_assign_parameter_uint_pointer    (char *file, uint line, cpar_table *vp,
                                                     char *key, uint **, uint type);
bool _cpar_assign_parameter_double_pointer  (char *file, uint line, cpar_table *vp,
                                                     char *key, double **, uint type);
bool _cpar_assign_parameter_real_pointer    (char *file, uint line, cpar_table *vp,
                                                     char *key, real **, uint type);
bool _cpar_assign_parameter_cpar_pointer    (char *file, uint line, cpar_table *vp,
                                                     char *key, cpar_table **, uint type);
bool _cpar_assign_parameter_gpointer        (char *file, uint line, cpar_table *vp,
                                                     char *key, gpointer *, uint type);

// Get
#define CPAR_GET_TYPE(vp, key) \
    _cpar_get_type                          (__FILE__, __LINE__, vp, key)
#define CPAR_GET_RAW(vp, key) \
    _cpar_get_raw                           (__FILE__, __LINE__, vp, key)
#define CPAR_GET_BOOL(vp, key) \
    _cpar_get_parameter_bool                (__FILE__, __LINE__, vp, key, CPAR_BOOL)
#define CPAR_GET_INT(vp, key) \
    _cpar_get_parameter_int                 (__FILE__, __LINE__, vp, key, CPAR_INT)
#define CPAR_GET_UINT(vp, key) \
    _cpar_get_parameter_uint                (__FILE__, __LINE__, vp, key, CPAR_UINT)
#define CPAR_GET_DOUBLE(vp, key) \
    _cpar_get_parameter_double              (__FILE__, __LINE__, vp, key, CPAR_DOUBLE)
#define CPAR_GET_REAL(vp, key) \
    _cpar_get_parameter_real                (__FILE__, __LINE__, vp, key, CPAR_REAL)
#define CPAR_GET_STRING(vp, key) \
    _cpar_get_parameter_string              (__FILE__, __LINE__, vp, key, CPAR_STRING)
#define CPAR_GET_BOOL_POINTER(vp, key) \
    _cpar_get_parameter_bool_pointer        (__FILE__, __LINE__, vp, key, CPAR_BOOL_POINTER)
#define CPAR_GET_CHAR_POINTER(vp, key) \
    _cpar_get_parameter_char_pointer        (__FILE__, __LINE__, vp, key, CPAR_CHAR_POINTER)
#define CPAR_GET_INT_POINTER(vp, key) \
    _cpar_get_parameter_int_pointer         (__FILE__, __LINE__, vp, key, CPAR_INT_POINTER)
#define CPAR_GET_UINT_POINTER(vp, key) \
    _cpar_get_parameter_uint_pointer        (__FILE__, __LINE__, vp, key, CPAR_UINT_POINTER)
#define CPAR_GET_DOUBLE_POINTER(vp, key) \
    _cpar_get_parameter_double_pointer      (__FILE__, __LINE__, vp, key, CPAR_DOUBLE_POINTER)
#define CPAR_GET_REAL_POINTER(vp, key) \
    _cpar_get_parameter_real_pointer        (__FILE__, __LINE__, vp, key, CPAR_REAL_POINTER)
#define CPAR_GET_CPAR_POINTER(vp, key) \
    _cpar_get_parameter_cpar_pointer        (__FILE__, __LINE__, vp, key, CPAR_CPAR_POINTER)
#define CPAR_GET_GPOINTER(vp, key, gpointer_type) \
    ((gpointer_type ## _TYPECAST()) _cpar_get_parameter_gpointer(__FILE__, __LINE__, vp, key, gpointer_type))
uint _cpar_get_type                     (char *file, uint line, cpar_table *vp, char *key);
gpointer     _cpar_get_raw                      (char *file, uint line, cpar_table *vp, char *key);
bool         _cpar_get_parameter_bool           (char *file, uint line, cpar_table *vp, char *key, 
                                                 uint type);
int          _cpar_get_parameter_int            (char *file, uint line, cpar_table *vp, char *key, 
                                                 uint type);
uint         _cpar_get_parameter_uint           (char *file, uint line, cpar_table *vp, char *key, 
                                                 uint type);
double       _cpar_get_parameter_double         (char *file, uint line, cpar_table *vp, char *key, 
                                                 uint type);
real         _cpar_get_parameter_real           (char *file, uint line, cpar_table *vp, char *key, 
                                                 uint type);
bool*        _cpar_get_parameter_bool_pointer   (char *file, uint line, cpar_table *vp, char *key, 
                                                 uint type);
char*        _cpar_get_parameter_string         (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
char*        _cpar_get_parameter_char_pointer   (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
int*         _cpar_get_parameter_int_pointer    (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
uint*        _cpar_get_parameter_uint_pointer   (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
double*      _cpar_get_parameter_double_pointer (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
real*        _cpar_get_parameter_real_pointer   (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
cpar_table*  _cpar_get_parameter_cpar_pointer   (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);
gpointer     _cpar_get_parameter_gpointer       (char *file, uint line, cpar_table *vp, char *key,
                                                 uint type);

// Table changes
#define CPAR_KEEP(vp, ...) \
        _cpar_keep_parameters_by_key   (__FILE__, __LINE__, vp, __VA_ARGS__, NULL)
#define CPAR_KEEP_MATCH(vp, string) \
        _cpar_keep_parameters_by_string(__FILE__, __LINE__, vp, string)
#define CPAR_LIST(vp) \
        _cpar_list_parameters   (__FILE__, __LINE__, vp)
#define CPAR_REMOVE(vp, ...) \
        _cpar_remove_parameters_by_key (__FILE__, __LINE__, vp, __VA_ARGS__, NULL)
#define CPAR_REMOVE_MATCH(vp, string) \
        _cpar_remove_parameters_by_string(__FILE__, __LINE__, vp, string)
#define CPAR_FREE(vp) \
        _cpar_destroy_cpar_table(__FILE__, __LINE__, vp)

#define CPAR_MERGE(vp1, vp2) \
        _cpar_merge_cpar_tables(__FILE__, __LINE__, vp1, vp2, CPAR_MERGE_NO_OVERWRITE_MODE)
#define CPAR_MERGE_OVERWRITE(vp1, vp2) \
        _cpar_merge_cpar_tables(__FILE__, __LINE__, vp1, vp2, CPAR_MERGE_OVERWRITE_MODE)
#define CPAR_MERGE_NEW(vp1, vp2) \
        _cpar_merge_cpar_tables(__FILE__, __LINE__, vp1, vp2, CPAR_MERGE_NEW_MODE)
#define CPAR_MERGE_UPDATE(vp1, vp2) \
        _cpar_merge_cpar_tables(__FILE__, __LINE__, vp1, vp2, CPAR_MERGE_UPDATE_MODE)

#define CPAR_UPDATE(vp, ...) \
        _cpar_add_parameters_macro(__FILE__, __LINE__, vp, CPAR_MERGE_NO_OVERWRITE_MODE, \
                                   __VA_ARGS__, CPAR_ADD_PARAMETERS_END)
#define CPAR_UPDATE_OVERWRITE(vp, ...) \
        _cpar_add_parameters_macro(__FILE__, __LINE__, vp, CPAR_MERGE_OVERWRITE_MODE, \
                                   __VA_ARGS__, CPAR_ADD_PARAMETERS_END)
#define CPAR_UPDATE_NEW(vp, ...) \
        _cpar_add_parameters_macro(__FILE__, __LINE__, vp, CPAR_MERGE_NEW_MODE, \
                                   __VA_ARGS__, CPAR_ADD_PARAMETERS_END)
#define CPAR_UPDATE_UPDATE(vp, ...) \
        _cpar_add_parameters_macro(__FILE__, __LINE__, vp, CPAR_MERGE_UPDATE_MODE, \
                                   __VA_ARGS__, CPAR_ADD_PARAMETERS_END)

void        _cpar_keep_parameters_by_key    (char *file, uint line, cpar_table *vp, ...);
void        _cpar_keep_parameters_by_string (char *file, uint line, cpar_table *vp, char *string);
void        _cpar_list_parameters           (char *file, uint line, cpar_table *vp);
void        _cpar_list_parameters_GHFunc    (gpointer key, gpointer value, gpointer user_data);
void        _cpar_remove_parameters_by_key  (char *file, uint line, cpar_table *vp, ...);
void       _cpar_remove_parameters_by_string(char *file, uint line, cpar_table *vp, char *string);
void        _cpar_destroy_cpar_table        (char *file, uint line, cpar_table *parameters_hash);
bool        _cpar_free_parameter            (gpointer key, gpointer value, gpointer user_data);
void        _cpar_merge_cpar_tables         (char *file, uint line, cpar_table *vp1, cpar_table *vp2, uint mode);
void        _cpar_merge_cpar_tables_GHFunc  (gpointer from_key_pointer, gpointer value, gpointer user_data);
void        _cpar_add_parameters_macro      (char *file, uint line, cpar_table *vp, 
                                             uint merge_mode, uint first_add_mode, ...);

// Flag values - passed to cpar_*_flag
#define CPAR_FLAG_SEEN      100
#define CPAR_FLAG_DEFAULT   101

// Flag actions
#define CPAR_FLAG_SET       200
#define CPAR_FLAG_CLEAR     201
#define CPAR_FLAG_ASSERT    202

// Tests
#define CPAR_EXISTS(vp, ...) \
        _cpar_exists            (__FILE__, __LINE__, vp, __VA_ARGS__, NULL)
#define CPAR_IS_ALL_SEEN(vp, verbose) \
        _cpar_is_all_flag_set   (__FILE__, __LINE__, vp, CPAR_FLAG_SEEN, verbose)
#define CPAR_IS_SEEN(vp, verbose, ...) \
        _cpar_is_flag_set       (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_SEEN, verbose, __VA_ARGS__, NULL)
#define CPAR_IS_SEEN_MATCH(vp, verbose, compare_string) \
        _cpar_is_flag_set_match (__FILE__, __LINE__, vp, CPAR_FLAG_SEEN, verbose, compare_string)
#define CPAR_IS_ALL_DEFAULT(vp, verbose) \
        _cpar_is_all_flag_set   (__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, verbose)
#define CPAR_IS_DEFAULT(vp, verbose, ...) \
        _cpar_is_flag_set       (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_DEFAULT, verbose, __VA_ARGS__, NULL)
#define CPAR_IS_DEFAULT_MATCH(vp, verbose, compare_string) \
        _cpar_is_flag_set_match (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_DEFAULT, verbose, compare_string)

bool _cpar_exists           (char *file, uint line, cpar_table *vp, ...);
bool _cpar_is_all_flag_set  (char *file, uint line, cpar_table *vp, uint flag, bool verbose);
bool _cpar_is_flag_set      (char *file, uint line, cpar_table *vp, uint flag, bool verbose, ...);
bool _cpar_is_flag_set_match(char *file, uint line, cpar_table *vp, uint flag, 
                             bool verbose, char *compare_string);

// Internal tests
bool _cpar_type_is_gpointer_custom(uint type);
bool _cpar_type_is_normal(uint type);
void _cpar_assign_pointer_type(char *file, uint line, char *type_variable_string, 
                               uint *type_variable);

// Seen flag
#define CPAR_CLEAR_ALL_SEEN(vp)     \
        _cpar_set_all_flag      (__FILE__, __LINE__, vp, CPAR_FLAG_SEEN, CPAR_FLAG_CLEAR)
#define CPAR_CLEAR_SEEN(vp, ...)    \
        _cpar_set_flag          (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_SEEN, CPAR_FLAG_CLEAR, __VA_ARGS__, NULL)
#define CPAR_CLEAR_SEEN_MATCH(vp, compare_string)    \
        _cpar_set_flag_match    (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_SEEN, CPAR_FLAG_CLEAR, compare_string)
#define CPAR_SET_ALL_SEEN(vp)       \
        _cpar_set_all_flag      (__FILE__, __LINE__, vp, CPAR_FLAG_SEEN, CPAR_FLAG_SET)
#define CPAR_SET_SEEN(vp, ...)      \
        _cpar_set_flag          (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_SEEN, CPAR_FLAG_SET, __VA_ARGS__, NULL)
#define CPAR_SET_SEEN_MATCH(vp, compare_string)      \
        _cpar_set_flag_match    (__FILE__, __LINE__, vp, \
                                 CPAR_FLAG_SEEN, CPAR_FLAG_SET, compare_string)

// Default flag
#define CPAR_CLEAR_ALL_DEFAULT(vp)  \
        _cpar_set_all_flag  (__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, CPAR_FLAG_CLEAR)
#define CPAR_CLEAR_DEFAULT(vp, ...) \
        _cpar_set_flag      (__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, CPAR_FLAG_CLEAR, \
                             __VA_ARGS__, NULL)
#define CPAR_CLEAR_DEFAULT_MATCH(vp, compare_string) \
        _cpar_set_flag_match(__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, CPAR_FLAG_CLEAR, \
                             compare_string)
#define CPAR_SET_ALL_DEFAULT(vp)    \
        _cpar_set_all_flag  (__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, CPAR_FLAG_SET)
#define CPAR_SET_DEFAULT(vp, ...)   \
        _cpar_set_flag      (__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, CPAR_FLAG_SET, \
                             __VA_ARGS__, NULL)
#define CPAR_SET_DEFAULT_MATCH(vp, compare_string)   \
        _cpar_set_flag_match(__FILE__, __LINE__, vp, CPAR_FLAG_DEFAULT, CPAR_FLAG_SET, \
                             compare_string)

void _cpar_set_all_flag     (char *file, uint line, cpar_table *vp, uint flag, uint action);
void _cpar_set_flag         (char *file, uint line, cpar_table *vp, uint flag, uint action, ...);
void _cpar_set_flag_match   (char *file, uint line, cpar_table *vp, 
                             uint flag, uint action, char *compare_string);


// Assertions
#define CPAR_ASSERT_CPAR_TABLE(vp) _cpar_assert_cpar_table(__FILE__, __LINE__, vp)
void _cpar_assert_type(char *file, uint line, char *function, uint type);
void _cpar_assert_cpar_table(char *file, uint line, cpar_table *vp);

// Custom gpointer stuff
#define CPAR_ASSIGN_POINTER_TYPE(type_variable) \
        _cpar_assign_pointer_type(__FILE__, __LINE__, #type_variable, &type_variable)

#define _cpar_abort(file, line, ...)  \
    _cpar_abort_function(__FILE__, __LINE__, CPAR_FUNCTION_NAME, file, line, __VA_ARGS__)

// Command-line processing
#define CPAR_PROCESS_COMMAND_LINE(vp, argc, argv, ...) \
        _cpar_process_command_line(__FILE__, __LINE__, vp, argc, argv, __VA_ARGS__, NULL)
void _cpar_process_command_line(char *file, uint line, cpar_table *vp, int argc, char *argv[], ...);

// This function is called if an unrecoverable error occurs.  It can be changed using _cpar_set_abort_function().
void (*_cpar_abort_function)(char *file_i, uint line_i, const char *function_i,
                             char *file, uint line, char *fmt, ...);
void cpar_set_abort_function(void (*new_abort_function)(char *file_i, uint line_i, const char *function_i, 
                                                        char *file, uint line, char *fmt, ...));
void _cpar_internal_abort_function(char *file_i, uint line_i, const char *function_i,
                                   char *file, uint line, char *fmt, ...);

// Misc functions
const char * _cpar_type_name(uint type);

// End #ifndef _C_PARAMETERS_H
#endif
